/* FILE: display.h                 -*-Mode: c++-*-
 *
 * C++ classes for display windows
 *
 */

#ifndef DISPLAY_H
#define DISPLAY_H

#include "nb.h"
#include "vf.h"

#include "bitmap.h"
#include "colormap.h"

/* End includes */     /* Optional directive to build.tcl */

//////////////////////////////////////////////////////////////////////////
// PlotConfiguration structure
//
class PlotConfiguration {
private:
  static const ClassDoc class_doc;
public:
  PlotConfiguration();
  PlotConfiguration(const PlotConfiguration &);
  PlotConfiguration& operator=(const PlotConfiguration &);
  
  BOOL displaystate; // 1=>display, 0=>don't display

  BOOL autosample;   // If 1, then replace specified subsample
  /// with an automatically determined data-dependent value.
  REAL4m subsample;  // Subsample period, in units of mesh cells.
  ///   NOTE: If subsample=0, then *all* points are selected.
  /// Also if subsample<1, then features will be scaled as
  /// though subsample==1.

  REAL4m size;   // Item magnification, relative to its
  ///  "standard size".  The drawn size depends upon size,
  /// subsample, and the "standard size".

  BOOL antialias;  // 1=> Do antialiasing, 0=> don't.  May be ignored
  /// by some drawing routines.

  const char* stipple; // Stipple pattern.  Use "" for solid.
  // "gray25" is recommended for partial transparency.
  // NOTE: The patterns are intended to be string constants,
  // so this element points at that string; no memory is
  // allocated for the pattern name.  If a client points
  // stipple into free store, it is the responsibility of
  // the client to guarantee that the pointer stays valid
  // through the lifetime of the PlotConfiguration object.

  Nb_DString colorquantity;

  DisplayColorMap colormap;
  const char* Color(REAL4m shade) const {
    return colormap.GetColor(shade);
  }
  OommfPackedRGB PackedColor(REAL4m shade) const {
    return colormap.GetPackedColor(shade);
  }

};

//////////////////////////////////////////////////////////////////////////
// Display frame class
//
//  Restrictions: Assumes square pixels.
//                2D, ignores z.
//
enum CoordinateSystem { // All are righthanded systems
  CS_ILLEGAL=-1,             // Illegal value
  CS_CALCULATION_STANDARD=0, // x to left,  y up,       z out -  mesh units
  CS_DISPLAY_STANDARD=1,     // x to left,  y down,     z in  - pixel units
  CS_DISPLAY_ROT_90=2,       // x down,     y to right, z in  - pixel units
  CS_DISPLAY_ROT_180=3,      // x to right, y up,       z in  - pixel units
  CS_DISPLAY_ROT_270=4       // x up,       y to left,  z in  - pixel units
};

INT4m CoordinateAngleDifference(CoordinateSystem cs1,CoordinateSystem cs2,
				double &angle);
// Calculates the difference (in degrees, from 0 to 360) of cs1 - cs2, if
// both are CS_DISPLAY_* coordinates, sets the difference in 'angle' and
// returns 0;  If cs1 is not a CS_DISPLAY_* c.s., then the return value
// is 1; if cs2 is not a CS_DISPLAY_* c.s., then the return value is 2.
// In these latter two cases angle is 0.0 on return.

struct SampleInfo {
  REAL4m subsample;
  Nb_DString colorquantity;
  Nb_List<Vf_DisplayVector> dvlist;
  SampleInfo() { subsample=0.; } 
};

class DisplayFrame
{
private:
  static const ClassDoc class_doc;
  static const REAL4m default_arrow_cellsize;
  static const REAL4m default_pixel_cellsize; 
  /// Preferred cellsizes, in pixels.

  static const REAL4m default_arrow_ratio;
  static const REAL4m default_pixel_ratio;
  /// Default (base) size of objects, relative to the cell size.

  Tcl_Interp* tcl_interp;
  Vf_Mesh *mesh;

  void GetBoundingBoxes(Nb_BoundingBox<REAL4> &bdry,
			Nb_BoundingBox<REAL4> &data,REAL8m &data_pad) const;
  /// NB: Either of the returning boxes, in particular bdry, may
  ///  be an "empty" box, in which case max<min.

  INT4m GetDisplayList(SampleInfo &si);

  REAL4m mesh_per_pixel;  // * pixels => mesh units;   This is
  ///  guaranteed>0.  User zoom is given by
  ///    mesh->GetApproximateCellWidth()/mesh_per_pixel;

  REAL4m preferred_arrow_cellsize,preferred_pixel_cellsize;

  CoordinateSystem coords; // Current coordinates
  void ChangeCoordinates(CoordinateSystem old_coords,
			 CoordinateSystem new_coords,
			 Nb_List<Vf_DisplayVector> &display_list) const;
  void ChangeCoordinates(CoordinateSystem old_coords,
			 CoordinateSystem new_coords,
			 Nb_List< Nb_Vec3<REAL4> > &pt_list) const;

  PlotConfiguration arrow_config,pixel_config;

  SampleInfo arrow_sample,pixel_sample;
  /// Current mesh sample list and support information.

  BOOL arrow_valid,pixel_valid,boundary_valid;
  /// If true, then the corresponding mesh sample lists are up-to-date.
  /// In general, display_lists are only updated when a Render request
  /// is made.  Exceptions are scaling and coordinate change requests,
  /// for which valid lists will be adjusted at once.

  void MarkAllSampleListsInvalid() {
    arrow_valid=pixel_valid=boundary_valid=0;
  }

  REAL8m GetPixelSize() const;
  REAL8m GetArrowSize() const;

  INT4m DrawArrow(const char *canvas_name,const Vf_DisplayVector &locvec,
                  REAL4m arrowsize,BOOL hide) const;
  INT4m DrawPixel(const char *canvas_name,const Vf_DisplayVector &locvec,
                  REAL4m pixelsize,const char* stipple,BOOL hide) const;

  BOOL draw_boundary;        // 1=>draw boundary, 0=>don't draw boundary
  Nb_List< Nb_Vec3<REAL4> > boundary_list; // List of boundary vertices
  INT4m DrawLine(const char *canvas_name,
		 const Nb_Vec3<REAL4> &Pt1,const Nb_Vec3<REAL4> &Pt2,
		 const char* tags);

  Nb_DString background_color; // Color in Tk 8-bit format, "#XXXXXX".

  DisplayFrame(DisplayFrame&);            // Disable copy constructor
  DisplayFrame& operator=(DisplayFrame&); // and assignment operator

public:
  DisplayFrame();

  void SetMesh(Vf_Mesh *newmesh);

  void SetTclInterp(Tcl_Interp *newinterp) { tcl_interp=newinterp; }

  void GetPlotConfiguration(PlotConfiguration& _arrow_config,
			    PlotConfiguration& _pixel_config) const {
    _arrow_config=arrow_config;
    _pixel_config=pixel_config;
  }
  void SetPlotConfiguration(const PlotConfiguration& _arrow_config,
			    const PlotConfiguration& _pixel_config);

  void SetPreferredArrowCellsize(REAL4m pixels);
  void SetPreferredPixelCellsize(REAL4m pixels);
  REAL4m GetAutoSampleRate(REAL4m cellsize) const;
  /// Returns recommended subsample rate, for the current mesh
  /// at current zoom.

  CoordinateSystem GetCoordinates() const { return coords; }
  BOOL SetCoordinates(CoordinateSystem new_coords);

  const char* GetBackgroundColor() { return background_color; }
  void SetBackgroundColor(const char* color);
  /// GetBackgroundColor returns a Tk 8-bit color spec, of the form
  /// "#XXXXXX".  SetBackgroundColor accepts any valid Tk color string,
  /// and converts it to the "#XXXXXX" format for internal use.

  BOOL GetDrawBoundary() const { return draw_boundary; }
  void SetDrawBoundary(BOOL newstate) { draw_boundary=newstate; }

  INT4m GetColorQuantityTypes(Nb_List<Nb_DString> &types) const {
    return mesh->ColorQuantityTypes(types);
  }

  const char* GetVecColor(const Nb_Vec3<REAL4>& v) const;
  /// Access to the underlying Vf_Mesh::GetVecShade function.  See that
  /// routine for details on shade determination.  The mapping from
  /// shade to color depends on the state of arrow_config.displaystate
  /// and pixel_config.displaystate.  Pixel config colorquantity and
  /// colormap are used, unless pixels are turned off and arrows are
  /// turned on, in which case arrrow config colorquantity and colormap
  /// are used.

  INT4m Render(const char *canvas_name,BOOL hide=0);
  // If hide is true, then pixel and arrow objects are created
  // with the '-state hidden' switch.  Note that this is only
  // available on Tk 8.3 and later.  If using an older version,
  // the default hide=0 should be selected.

  INT4m FillBitmap(OommfBitmap& bitmap);
  /// Writes output into bitmap

  REAL4m GetZoom(void) const;
  REAL4m SetZoom(REAL4m zoom);
  REAL4m SetZoom(REAL4m width,REAL4m height);
  /// NOTE: "zoom" is in units of "pixels per approx cell size".

  void GetRequestedSubsampleRates(REAL4m &arrow_samplerate,
				  REAL4m &pixel_samplerate) const {
    arrow_samplerate=arrow_config.subsample;
    pixel_samplerate=pixel_config.subsample;
  }
  void GetActualSubsampleRates(REAL4m &arrow_samplerate,
			       REAL4m &pixel_samplerate) const {
    arrow_samplerate=arrow_sample.subsample;
    pixel_samplerate=pixel_sample.subsample;
  }

  Nb_BoundingBox<REAL4> GetDisplayBox() const;
  /// Returns extents of mesh, in pixels (at current zoom).

  const char* GetValueUnit() const { return mesh->GetValueUnit(); }

  REAL8m GetDefaultValueScaling() const;
  REAL8m GetValueScaling() const;
  INT4m SetValueScaling(REAL8m newscaling);

};

#endif // DISPLAY_H

