/* FILE: ouput.h                 -*-Mode: c++-*-
 *
 * Output classes.  Each class providing output should embed an
 * instance of an output class for each output provided.
 *
 * NOTE: This code looks pretty fragile, because there is a lot
 *  of coupling between the output classes and their owner class.
 *  Maybe we can figure out a cleaner, more robust solution in
 *  the future.
 */

#ifndef _OXS_OUTPUT
#define _OXS_OUTPUT

#include "oc.h"
#include "simstate.h"
#include "oxsexcept.h"

OC_USE_STD_NAMESPACE;  // Specify std namespace, if supported

/* End includes */

class Oxs_Director;   // Forward declaration


class Oxs_Output {
private:
  Oxs_Director* director;  // Director instance this Oxs_Output
  /// instance is registered with.

  string owner_name;
  string output_name;
  string output_type;
  string output_units;

  const BOOL cache_support; // If true, then caching is supported
  UINT4m cache_request_count;

  // Disable copy constructor by declaring it without defining it.
  Oxs_Output(const Oxs_Output&);

protected:
  BOOL ForceCacheIncrement(INT4m incr);
  /// Analogous to CacheRequestIncrement, but ignores
  /// cache_support value.  Returns 1 if increment was
  /// successful.  For use by child classes.

public:
  Oxs_Output(const char* owner_name_,
	     const char* output_name_,
	     const char* output_type_,
	     const char* output_units_,
	     BOOL cache_support_,
	     Oxs_Director* director_=NULL);
  /// If director is not NULL, then construction automatically registers
  /// with director_, and automatic deregistration is set up in
  /// ~Oxs_Output.
  virtual ~Oxs_Output(); // Deregisters

  string OwnerName()  const { return owner_name;  }
  string OutputName() const { return output_name; }
  string OutputType() const { return output_type; }
  string OutputUnits() const { return output_units; }

  virtual int Output(const Oxs_SimState*,Tcl_Interp*,int argc,char** argv)
    =0;

  // Cache flag control.  Returns 1 if increment was successful.
  BOOL CacheRequestIncrement(INT4m incr)
  {
    if(!cache_support) return 0; // Caching disabled
    return ForceCacheIncrement(incr);
  }

  UINT4m GetCacheRequestCount() const { return cache_request_count; }
};

//////////////////////////////////////////////////////////////////
// Templated cache class, used by templated output classes.
template<class T> struct Oxs_OutputCache {
public:
  Oxs_OutputCache() : state_id(0) {}
  UINT4m state_id;
  T value;
};

//////////////////////////////////////////////////////////////////
// Templated output classes.  The templated class T should be the
// embedding class.

////////////////////////
// Scalar output
template<class T> class Oxs_ScalarOutput:public Oxs_Output {
  typedef void (T::* GetScalar)(const Oxs_SimState&);
private:
  T* owner;
  const GetScalar getscalar;
  /// Without the typedef, this would be
  ///    void (T::* const getscalar)(const Oxs_SimState&)

  // Disable copy constructor by declaring it without defining it.
  Oxs_ScalarOutput(const Oxs_ScalarOutput<T> &);

public:
  Oxs_ScalarOutput(T* owner_,
		   const char* owner_name_,const char* output_name_,
		   const char* output_units_,
		   BOOL cache_support_,const GetScalar member_function,
		   Oxs_Director* director_=NULL)
    : Oxs_Output(owner_name_,output_name_,"scalar",output_units_,
		 cache_support_,director_),
      owner(owner_),
      getscalar(member_function)
  {
    cache.value=0; // Safety
    CacheRequestIncrement(1);  // Temporary measure: until /**/
    /// cache management is implemented in Tcl interface,  /**/
    /// enable caching on all scalar outputs.              /**/
  }
  // Note 1: The import owner_ is used to set the data member owner,
  //  but is not otherwised used inside the Oxs_ScalarOutput constructor.
  //  In particular, this means that the calling T object can safely pass
  //  in its 'this' pointer from its constructor.
  // Note 2: If the implementation of this function is given outside
  //  the class definition body, then Microsoft VC++ 5.0 chokes on the
  //  pointer-to-member-function T::*, and dies with "internal compiler
  //  error."

  // Public data member for result drop-off/storage.
  // The owner should test GetCacheRequestCount() inside the output
  // calculation code, and adjust the cache as needed.
  Oxs_OutputCache<REAL8m> cache;

  virtual int Output(const Oxs_SimState* state,Tcl_Interp* interp,
		     int argc,char** argv);
  // The owner getscalar() function can assume on entry that the
  // cache value stored here is invalid, but it may become valid
  // during the lifetime of the getscalar() call (if getscalar()
  // calls some other function that sets the cache).
};

////////////////////////
// Vector field output
template<class T> class Oxs_VectorFieldOutput:public Oxs_Output {
  typedef void (T::* GetVectorField)(const Oxs_SimState&);
private:
  T* owner;
  const GetVectorField getvectorfield;
  /// Without the typedef, this would be
  ///    void (T::* const getvectorfield)(const Oxs_SimState&)

  // Disable copy constructor by declaring it without defining it.
  Oxs_VectorFieldOutput(const Oxs_VectorFieldOutput<T> &);

public:
  Oxs_VectorFieldOutput(T* owner_,
		   const char* owner_name_,const char* output_name_,
  		   const char* output_units_,
		   BOOL cache_support_,
		   const GetVectorField member_function,
		   Oxs_Director* director_=NULL)
    : Oxs_Output(owner_name_,output_name_,"vector field",output_units_,
		 cache_support_,director_),
      owner(owner_),
      getvectorfield(member_function)
  {}
  // Note 1: The import owner_ is used to set the data member owner,
  //  but is not otherwised used inside the Oxs_VectorFieldOutput
  //  constructor.  In particular, this means that the calling T
  //  object can safely pass in its 'this' pointer from its constructor.
  // Note 2: If the implementation of this function is given outside
  //  the class definition body, then Microsoft VC++ 5.0 chokes on the
  //  pointer-to-member-function T::*, and dies with "internal compiler
  //  error."

  // Public data member for result drop-off/storage.
  // The owner should test GetCacheRequestCount() inside the output
  // calculation code, and adjust the cache as needed.
  Oxs_OutputCache<Oxs_MeshValue<ThreeVector> > cache;

  virtual int Output(const Oxs_SimState* state,Tcl_Interp* interp,
		     int argc,char** argv);
};

////////////////////////////////////////////////////////////////////////
// Include definitions below.  This is required by compiler/linkers
// using what gcc calls the "Borland model," as opposed to the "Cfront
// model."  The latter is distinguished by the use of template
// repositories.  With the Cfront model, template definitions can be
// put in a file (*.cc) separate from the header file (*.h).  Putting
// the definitions in the header file along with declarations slows down
// compilation, because the definitions are re-compiled in each including
// file, but appears to be more portable.

//////////////////////////////////////////////////////////////
// Oxs_ScalarOutput<T> definitions
template<class T>
int Oxs_ScalarOutput<T>::Output
(const Oxs_SimState* state,   // State to get output for
 Tcl_Interp* interp,          // Tcl interpreter
 int argc,char** /* argv */)  // Args
{
  Tcl_ResetResult(interp);
  if(argc!=0) {
    Tcl_AppendResult(interp, "wrong # of args: should be none",
		     (char *) NULL);
    return TCL_ERROR;
  }
  if(state==NULL || state->Id()==0) {
    Tcl_AppendResult(interp,"Invalid state passed to output object",
		     (char *)NULL);
    return TCL_ERROR;
  }

  if(cache.state_id!=state->Id()) {
    // Fill cache with vector field
    ForceCacheIncrement(1);
    (owner->*getscalar)(*state);
    ForceCacheIncrement(-1);

    // Check that cache got filled
    if(cache.state_id!=state->Id()) {
      OC_THROW(Oxs_ProgramLogicError,
	       "Output cache not filled by registered getscalar"
	       " call (as required).");
    }
  }

  static char buf[100];
  sprintf(buf,"%.17g",cache.value);
  Tcl_AppendResult(interp,buf,(char *)NULL);
  return TCL_OK;
}


//////////////////////////////////////////////////////////////
// Oxs_VectorFieldOutput<T> definitions
template<class T>
int Oxs_VectorFieldOutput<T>::Output
(const Oxs_SimState* state, // State to get output for
 Tcl_Interp* interp,        // Tcl interpreter
 int argc,char** argv)      // Args
{
  Tcl_ResetResult(interp);
  if(argc!=1) {
    Tcl_AppendResult(interp, "wrong # of args: should be 1: filename",
		     (char *) NULL);
    return TCL_ERROR;
  }
  if(state==NULL || state->Id()==0) {
    Tcl_AppendResult(interp,"Invalid state passed to output object",
		     (char *)NULL);
    return TCL_ERROR;
  }

  char* filename = argv[0];

  if(cache.state_id!=state->Id()) {
    // Fill cache with vector field
    ForceCacheIncrement(1);
    cache.state_id=0;
    cache.value.AdjustSize(state->mesh);
    (owner->*getvectorfield)(*state);
    ForceCacheIncrement(-1);

    // Check that cache got filled
    if(cache.state_id!=state->Id()) {
      OC_THROW(Oxs_ProgramLogicError,
	       "Output cache not filled by registered getvectorfield"
	       " call (as required).");
    }
  }

  // Write cache to file
  state->mesh->WriteOvf(filename,
			"Sample Oxs_VectorFieldOutput<T> output",
			"test","rectangular","binary","8",
			&cache.value,NULL);

  Tcl_AppendResult(interp,filename,(char *)NULL);
  return TCL_OK;
}


#endif // _OXS_OUTPUT
