/* FILE: demag.cc            -*-Mode: c++-*-
 *
 * Average H demag field across rectangular cells.  This is a modified
 * version of the simpledemag class, which uses symmetries in the
 * interaction coefficients to reduce memory usage.
 *
 */

#include "nb.h"
#include "director.h"
#include "key.h"
#include "mesh.h"
#include "meshvalue.h"
#include "simstate.h"
#include "threevector.h"
#include "energy.h"		// Needed to make MSVC++ 5 happy

#include "rectangularmesh.h"
#include "demag.h"
#include "demagcoef.h"
#include "fft.h"

// Oxs_Ext registration support
OXS_EXT_REGISTER(Oxs_Demag);

/* End includes */

// Helper function
UINT4m Oxs_Demag::NextPowerOfTwo(UINT4m n) const
{ // Returns first power of two >= n
  UINT4m m=1;
  while(m<n) m*=2;
  return m;
}

// Constructor
Oxs_Demag::Oxs_Demag(
  const char* name,     // Child instance id
  Oxs_Director* newdtr, // App director
  Tcl_Interp* safe_interp, // Safe interpreter
  const char* argstr)   // MIF input block parameters
  : Oxs_Energy(name,newdtr,safe_interp,argstr),
    xdim(0),ydim(0),zdim(0),totalsize(0),
    pxdim(0),pydim(0),pzdim(0),ptotalsize(0),mesh_id(0),
    A00(NULL),A01(NULL),A02(NULL),A11(NULL),A12(NULL),A22(NULL),
    xcomp(NULL),ycomp(NULL),zcomp(NULL)
{
  VerifyAllInitArgsUsed();
}

void Oxs_Demag::ReleaseMemory()
{
  if(xcomp!=NULL) { delete[] xcomp; xcomp=NULL; }
  if(ycomp!=NULL) { delete[] ycomp; ycomp=NULL; }
  if(zcomp!=NULL) { delete[] zcomp; zcomp=NULL; }
  if(A00!=NULL) { delete[] A00; A00=NULL; }
  if(A01!=NULL) { delete[] A01; A01=NULL; }
  if(A02!=NULL) { delete[] A02; A02=NULL; }
  if(A11!=NULL) { delete[] A11; A11=NULL; }
  if(A12!=NULL) { delete[] A12; A12=NULL; }
  if(A22!=NULL) { delete[] A22; A22=NULL; }
  xdim=ydim=zdim=totalsize=0;
  pxdim=pydim=pzdim=ptotalsize=0;
}

void Oxs_Demag::FillCoefficientArrays(const Oxs_Mesh* genmesh) const
{ // This routine is conceptually const.
  const Oxs_RectangularMesh* mesh
    = dynamic_cast<const Oxs_RectangularMesh*>(genmesh);
  if(mesh==NULL) {
    string msg=string("Object ")
      + string(genmesh->InstanceName())
      + string(" is not a rectangular mesh.");
    throw Oxs_Ext::Error(msg.c_str());
  }

  // Fill dimension variables
  xdim = mesh->DimX();
  ydim = mesh->DimY();
  zdim = mesh->DimZ();
  totalsize=xdim*ydim*zdim;
  if(xdim==0 || ydim==0 || zdim==0) return; // Empty mesh!
  if(totalsize < xdim || totalsize < ydim || totalsize < zdim) {
    // Partial overflow check
    string msg="UINT4m overflow in " + string(InstanceName())
      + ": Product xdim*ydim*zdim too big to fit in a UINT4m variable";
    throw Oxs_Ext::Error(msg.c_str());
  }
  pxdim = NextPowerOfTwo(2*xdim);
  pydim = NextPowerOfTwo(2*ydim);
  pzdim = NextPowerOfTwo(2*zdim);
  ptotalsize=pxdim*pydim*pzdim;
  if(ptotalsize < pxdim || ptotalsize < pydim || ptotalsize < pzdim) {
    // Partial overflow check
    string msg="UINT4m overflow in " + string(InstanceName())
      + ": Product pxdim*pydim*pzdim too big to fit in a UINT4m variable";
    throw Oxs_Ext::Error(msg.c_str());
  }

  // Allocate memory for interaction matrices and magnetization components
  UINT4m axdim=1+pxdim/2,aydim=1+pydim/2,azdim=1+pzdim/2;
  UINT4m atotalsize=axdim*aydim*azdim;
  A00   = new REAL8m[atotalsize];
  A01   = new REAL8m[atotalsize];
  A02   = new REAL8m[atotalsize];
  A11   = new REAL8m[atotalsize];
  A12   = new REAL8m[atotalsize];
  A22   = new REAL8m[atotalsize];
  xcomp = new MyComplex[ptotalsize];
  ycomp = new MyComplex[ptotalsize];
  zcomp = new MyComplex[ptotalsize];
  if(xcomp==NULL || ycomp==NULL || zcomp==NULL
     || A00==NULL || A01==NULL || A02==NULL
     || A11==NULL || A12==NULL || A22==NULL) {
    // Safety check for those machines on which new[] doesn't throw
    // BadAlloc.
    string msg="Insufficient memory in cubicdemag constructor.";
    throw Oxs_Ext::Error(msg.c_str());
  }

  // According (16) in Newell's paper, the demag field is given by
  //                        H = -N*M
  // where N is the "demagnetizing tensor," with components Nxx, Nxy,
  // etc.  With the '-1' in 'scale' we store '-N' instead of 'N',
  // so we don't have to multiply the output from the FFT + iFFT
  // by -1 in ConstMagField() below.

  // Fill interaction matrices with demag coefs from Newell's paper.
  // Note that A00, A11 and A22 are even in x,y and z.
  // A01 is odd in x and y, even in z.
  // A02 is odd in x and z, even in y.
  // A12 is odd in y and z, even in x.
  // We use these symmetries to reduce storage requirements.  If
  // f is real and even, then f^ is also real and even.  If f
  // is real and odd, then f^ is (pure) imaginary and odd.
  // As a result, the transform of each of the Axx interaction
  // matrices will be real, with the same even/odd properties.
  //  To reduce computation, we pack an Axx pair together before
  // computing the FFT.  For example, we let xcomp = A00 + i.A01.
  // Then the FFT of xcomp will be A00^ + i.A01^, where both A00^
  // and A01^ are real.  Similarly, we pack A02 and A11 into ycomp,
  // and A12 and A22 into zcomp.

  UINT4m aindex,pindex,i,j,k;
  UINT4m pxydim=pxdim*pydim;
  UINT4m axydim=axdim*aydim;

  REALWIDE dx = mesh->EdgeLengthX();
  REALWIDE dy = mesh->EdgeLengthY();
  REALWIDE dz = mesh->EdgeLengthZ();
  // For demag calculation, all that matters is the relative
  // sizes of dx, dy and dz.  To help insure we don't run
  // outside floating point range, rescale these values so
  // largest is 1.0
  REALWIDE maxedge=dx;
  if(dy>maxedge) maxedge=dy;
  if(dz>maxedge) maxedge=dz;
  dx/=maxedge; dy/=maxedge; dz/=maxedge;

  REALWIDE scale = -1./(4*PI*dx*dy*dz);

  for(pindex=0;pindex<ptotalsize;pindex++) {
    // Initialize temp space to 0
    xcomp[pindex].Set(0.,0.);
    ycomp[pindex].Set(0.,0.);
    zcomp[pindex].Set(0.,0.);
  }

  REALWIDE a00,a01,a02,a11,a12,a22;

  for(k=0;k<zdim;k++) for(j=0;j<ydim;j++) for(i=0;i<xdim;i++) {
    REALWIDE x = dx*i;    REALWIDE y = dy*j;    REALWIDE z = dz*k;
    a00=scale*CalculateSDA00(x,y,z,dx,dy,dz);
    a01=scale*CalculateSDA01(x,y,z,dx,dy,dz);
    a02=scale*CalculateSDA02(x,y,z,dx,dy,dz);
    a11=scale*CalculateSDA11(x,y,z,dx,dy,dz);
    a12=scale*CalculateSDA12(x,y,z,dx,dy,dz);
    a22=scale*CalculateSDA22(x,y,z,dx,dy,dz);

    pindex = i+j*pxdim+k*pxydim;
    xcomp[pindex].Set(a00,a01);
    ycomp[pindex].Set(a02,a11);
    zcomp[pindex].Set(a12,a22);

    if(i>0) {
      pindex = (pxdim-i)+j*pxdim+k*pxydim;
      xcomp[pindex].Set(a00,-a01);
      ycomp[pindex].Set(-a02,a11);
      zcomp[pindex].Set(a12,a22);
    }
    if(j>0) {
      pindex = i+(pydim-j)*pxdim+k*pxydim;
      xcomp[pindex].Set(a00,-a01);
      ycomp[pindex].Set(a02,a11);
      zcomp[pindex].Set(-a12,a22);
    }
    if(k>0) {
      pindex = i+j*pxdim+(pzdim-k)*pxydim;
      xcomp[pindex].Set(a00,a01);
      ycomp[pindex].Set(-a02,a11);
      zcomp[pindex].Set(-a12,a22);
    }
    if(i>0 && j>0) {
      pindex = (pxdim-i)+(pydim-j)*pxdim+k*pxydim;
      xcomp[pindex].Set(a00,a01);
      ycomp[pindex].Set(-a02,a11);
      zcomp[pindex].Set(-a12,a22);
    }
    if(i>0 && k>0) {
      pindex = (pxdim-i)+j*pxdim+(pzdim-k)*pxydim;
      xcomp[pindex].Set(a00,-a01);
      ycomp[pindex].Set(a02,a11);
      zcomp[pindex].Set(-a12,a22);
    }
    if(j>0 && k>0) {
      pindex = i+(pydim-j)*pxdim+(pzdim-k)*pxydim;
      xcomp[pindex].Set(a00,-a01);
      ycomp[pindex].Set(-a02,a11);
      zcomp[pindex].Set(a12,a22);
    }
    if(i>0 && j>0 && k>0) {
      pindex = (pxdim-i)+(pydim-j)*pxdim+(pzdim-k)*pxydim;
      xcomp[pindex].Set(a00,a01);
      ycomp[pindex].Set(a02,a11);
      zcomp[pindex].Set(a12,a22);
    }
  }

  // Transform into frequency domain.
  fft.Forward(pxdim,pydim,pzdim,xcomp);
  fft.Forward(pxdim,pydim,pzdim,ycomp);
  fft.Forward(pxdim,pydim,pzdim,zcomp);

  // Copy results from xcomp into Axx.  We only need store 1/8th
  // of the results because of symmetries.
  for(k=0;k<azdim;k++) for(j=0;j<aydim;j++) for(i=0;i<axdim;i++) {
    aindex = i+j*axdim+k*axydim;
    pindex = i+j*pxdim+k*pxydim;
    A00[aindex] = xcomp[pindex].real();
    A01[aindex] = xcomp[pindex].imag();
    A02[aindex] = ycomp[pindex].real();
    A11[aindex] = ycomp[pindex].imag();
    A12[aindex] = zcomp[pindex].real();
    A22[aindex] = zcomp[pindex].imag();
  }
}

void Oxs_Demag::GetEnergyAndField
(const Oxs_SimState& state,
 Oxs_MeshValue<REAL8m>& energy,
 Oxs_MeshValue<ThreeVector>& field
 ) const
{
  // (Re)-initialize mesh coefficient array if mesh has changed.
  if(mesh_id != state.mesh->Id()) {
    mesh_id = 0; // Safety
    FillCoefficientArrays(state.mesh);
    mesh_id = state.mesh->Id();
  }

  const Oxs_MeshValue<ThreeVector>& spin = state.spin;
  const Oxs_MeshValue<REAL8m>& Ms = *(state.Ms);

  // Calculate FFT of Mx, My and Mz
  UINT4m xydim=xdim*ydim;
  UINT4m pxydim=pxdim*pydim;
  UINT4m index,pindex;
  UINT4m i,j,k;
  for(pindex=0;pindex<ptotalsize;pindex++) xcomp[pindex].Set(0.,0.);
  for(pindex=0;pindex<ptotalsize;pindex++) ycomp[pindex].Set(0.,0.);
  for(pindex=0;pindex<ptotalsize;pindex++) zcomp[pindex].Set(0.,0.);
  for(k=0;k<zdim;k++) for(j=0;j<ydim;j++) for(i=0;i<xdim;i++) {
    index  = i+j*xdim+k*xydim;
    pindex = i+j*pxdim+k*pxydim;
    REAL8m scale = Ms[index];
    const ThreeVector& vec = spin[index];
    xcomp[pindex].Set(scale*(vec.x),0);
    ycomp[pindex].Set(scale*(vec.y),0);
    zcomp[pindex].Set(scale*(vec.z),0);
  }
  fft.Forward(pxdim,pydim,pzdim,xcomp);
  fft.Forward(pxdim,pydim,pzdim,ycomp);
  fft.Forward(pxdim,pydim,pzdim,zcomp);
  
  // Calculate field components in frequency domain, then to iFFT to
  // transform back to space domain.  Make use of realness and even/odd
  // properties of interaction matrices Axx.
  UINT4m axdim=1+pxdim/2,aydim=1+pydim/2,azdim=1+pzdim/2;
  UINT4m axydim=axdim*aydim;
  UINT4m ai,aj,ak,akindex,ajindex,aindex;
  int isign,jsign,ksign;

  // Temp store
  MyComplex Mx,My,Mz;
  REAL8m a00,a11,a22,a01,a02,a12;

  pindex=0;
  for(k=0;k<pzdim;k++) {
    if(k<azdim) { ak=k;         ksign= 1; }
    else        { ak=pzdim-k;   ksign=-1; }
    akindex=ak*axydim;
    for(j=0;j<pydim;j++) {
      if(j<aydim) { aj=j;         jsign= 1; }
      else        { aj=pydim-j;   jsign=-1; }
      ajindex=akindex+aj*axdim;
      for(i=0;i<pxdim;i++) {
	if(i<axdim) { ai=i;         isign= 1; }
	else        { ai=pxdim-i;   isign=-1; }
	aindex=ajindex+ai;

	Mx = xcomp[pindex]; My = ycomp[pindex];	Mz = zcomp[pindex];
	a00 = A00[aindex];  a11 = A11[aindex];  a22 = A22[aindex];
	a01 = isign*jsign*A01[aindex];
	a02 = isign*ksign*A02[aindex];
	a12 = jsign*ksign*A12[aindex];

	xcomp[pindex] = a00*Mx + a01*My + a02*Mz;  // Hx
	ycomp[pindex] = a01*Mx + a11*My + a12*Mz;  // Hy
	zcomp[pindex] = a02*Mx + a12*My + a22*Mz;  // Hz

	++pindex;
      }
    }
  }
  fft.Inverse(pxdim,pydim,pzdim,xcomp);
  fft.Inverse(pxdim,pydim,pzdim,ycomp);
  fft.Inverse(pxdim,pydim,pzdim,zcomp);
  for(k=0;k<zdim;k++) for(j=0;j<ydim;j++) for(i=0;i<xdim;i++) {
    index  = i+j*xdim+k*xydim;
    pindex = i+j*pxdim+k*pxydim;
    field[index].x=xcomp[pindex].real();
    field[index].y=ycomp[pindex].real();
    field[index].z=zcomp[pindex].real();
  }

  // Calculate pointwise energy density: -0.5*MU0*<M,H>
  REAL8m mult = -0.5 * MU0;
  for(index=0;index<totalsize;index++) {
    energy[index] = mult * Ms[index] * (spin[index]*field[index]);
  }

}
