/* FILE: sectionatlas.cc                 -*-Mode: c++-*-
 *
 * Section atlas class, derived from OXS extension class.
 *
 */

#include "sectionatlas.h"

// Oxs_Ext registration support
OXS_EXT_REGISTER(Oxs_SectionAtlas);

/* End includes */


// Constructor
Oxs_SectionAtlas::Oxs_SectionAtlas
( const char* name,     // Child instance id
  Oxs_Director* newdtr, // App director
  Tcl_Interp* safe_interp, // Safe interpreter
  const char* argstr    // MIF block argument string
  ) : Oxs_Atlas(name,newdtr)
{
  // Use Tcl_SplitList to break argstr into components
  Oxs_SplitList arglist;
  if(arglist.Split(argstr)!=TCL_OK) {
    // Input string not a valid Tcl list; Throw exception
    string msg= string("MIF input block for ")
      + string(InstanceName())
      + string(" not a valid Tcl list: ")
      + string(argstr);
    throw Oxs_Ext::Error(msg.c_str());
  }
  if(arglist.Count()%2 != 0) {
    // Input string is not a list of pairs; Throw exception
    string msg=string("MIF input block for ")
      + string(InstanceName())
      + string(" does not have an even number of elements: ")
      + string(argstr);
    throw Oxs_Ext::Error(msg.c_str());
  }

  // Each pair is a name + section init string
  Oxs_Box bbox,tbox;
  BOOL expand_results = 0;
  Oxs_SplitList params;
  int i;
  for(i=0;i+1<arglist.Count();i+=2) {
    const string& name = arglist[i];
    const string& section_init_str = arglist[i+1];

    // Skip "ignore" name-section pairs
    if(name.compare("ignore")==0) continue;

    // Check that name isn't already used
    for(int j=0;j<section_name_list.size();j++) {
      if(name.compare(section_name_list[j])==0) {
	string msg = string("MIF input block for ")
	  + string(InstanceName())
	  + string(", subblock ") + name
	  + string(" invalid: section name ")
	  + name + string(" is repeated.");
	throw Oxs_Ext::Error(msg.c_str());
      }
    }

    // Otherwise pass section init string to Oxs_Ext constructor
    if(params.Split(section_init_str.c_str())!=TCL_OK
       || params.Count()!=2) {
      string msg = string("MIF input block for ")
	+ string(InstanceName())
	+ string(", subblock ") + name
	+ string(" improperly specified; parameter list should be a"
		 " section specifier + init string pair.");
      throw Oxs_Ext::Error(msg.c_str());
    }
    Oxs_Ext* tempext = MakeNew(params[0],
			       director,safe_interp,
			       params[1]);
    if(tempext==NULL) {
      string msg = "MIF input block for " + string(InstanceName())
	+ string(", subblock ") + name
	+ string(": Unable to create Oxs_Ext object ")
	+ params[0] + string(" with args ") + params[1];
      throw Oxs_Ext::Error(msg.c_str());
    }

    Oxs_Section* tr = dynamic_cast<Oxs_Section*>(tempext);
    if(tr==NULL) {
      string msg=string("Oxs_Ext object ") + params[0]
	+ string(" associated with input block ")
	+ string(InstanceName())
	+ string(", subblock ") + name
	+ string(", is not an Oxs_Section object");
      throw Oxs_Ext::Error(msg.c_str());
    }

    section_list.push_back(tr);
    section_name_list.push_back(name);

    tr->GetExtents(tbox);
    expand_results = bbox.Expand(tbox);
  }

  // Check that last section bounding box contain all others, by
  // comparing extent boxes. This doesn't find all containment
  // errors, but will at least catch some of them.
  if(arglist.Count()>2 && expand_results) {
      string msg=string("Final section ")
	+ string(arglist[arglist.Count()-2])
	+ string(" specified in input block ")
	+ string(InstanceName())
	+ string(" does not contain all preceding sections.");
      throw Oxs_Ext::Error(msg.c_str());
  }
}


// Destructor
Oxs_SectionAtlas::~Oxs_SectionAtlas()
{
  UINT4m count = UINT4m(section_list.size());
  for(UINT4m i=0;i<count;i++) {
    delete section_list[i];
  }
}

void Oxs_SectionAtlas::GetWorldExtents(Oxs_Box &mybox) const
{ // Fills mybox with bounding box for the atlas
  UINT4m world_index = UINT4m(section_list.size()) - 1;
  section_list[world_index]->GetExtents(mybox);
}

INT4m Oxs_SectionAtlas::GetRegionId(const ThreeVector& point) const
{ // Returns the id number for the first section in
  // "sections" list containing point.  The return
  // value is -1 if the point is not contained in
  // the atlas.
  UINT4m count = UINT4m(section_list.size());
  for(UINT4m i=0;i<count;i++) {
    if(section_list[i]->IsIn(point)) return i;
  }
  return -1;
}

INT4m Oxs_SectionAtlas::GetRegionId(const string& name) const
{ // Given a region id string (name), returns
  // the corresponding region id index.  If
  // "name" is not included in the atlas, then
  // -1 is returned.
  UINT4m count = UINT4m(section_list.size());
  for(UINT4m i=0;i<count;i++) {
    if(section_name_list[i].compare(name)==0) return i;
  }
  return -1;
}

INT4m Oxs_SectionAtlas::GetRegionName(UINT4m id,string& name) const
{ // Given an id number, fills in "name" with
  // the corresponding region id string.  Returns
  // 0 on success, 1 if id is invalid.
  if(id>=section_list.size()) return 1;
  name = section_name_list[id];
  return 0;
}
